/*
 * Copyright (c) 2004, PostgreSQL Global Development Group
 * See the LICENSE file in the project root for more information.
 */

package org.postgresql.jdbc2.optional;

/**
 * @deprecated Since 42.0.0, see {@link org.postgresql.ds.PGPoolingDataSource}
 */
@Deprecated
public class PoolingDataSource extends org.postgresql.ds.PGPoolingDataSource {
}
