/*
 * Copyright (c) 2020, PostgreSQL Global Development Group
 * See the LICENSE file in the project root for more information.
 */

/* @DefaultQualifier(value = NonNull.class, locations = TypeUseLocation.FIELD) */
/* @DefaultQualifier(value = NonNull.class, locations = TypeUseLocation.PARAMETER) */
/* @DefaultQualifier(value = NonNull.class, locations = TypeUseLocation.RETURN) */
package org.postgresql;

// import org.checkerframework.checker.nullness.qual.NonNull;
// import org.checkerframework.framework.qual.DefaultQualifier;
// import org.checkerframework.framework.qual.TypeUseLocation;
