/*
 * Copyright (c) 2018, PostgreSQL Global Development Group
 * See the LICENSE file in the project root for more information.
 */

package org.postgresql.util;

// import org.checkerframework.checker.nullness.qual.NonNull;
// import org.checkerframework.checker.nullness.qual.Nullable;

public interface Gettable<K extends /* @NonNull */ Object, V extends /* @NonNull */ Object> {
  /* @Nullable */ V get(K key);
}
