/* Automatically generated by GNU msgfmt.  Do not modify!  */
package org.postgresql.translation;
public class messages_bg extends java.util.ResourceBundle {
  private static final java.lang.String[] table;
  static {
    java.lang.String[] t = new java.lang.String[890];
    t[0] = "";
    t[1] = "Project-Id-Version: JDBC Driver for PostgreSQL 8.x\nReport-Msgid-Bugs-To: \nPO-Revision-Date: 2009-12-28 00:01+0100\nLast-Translator: <usun0v@mail.bg>\nLanguage-Team: <bg@li.org>\nLanguage: \nMIME-Version: 1.0\nContent-Type: text/plain; charset=UTF-8\nContent-Transfer-Encoding: 8bit\nX-Poedit-Language: Bulgarian\nX-Poedit-Country: BULGARIA\n";
    t[2] = "A CallableStatement function was executed and the out parameter {0} was of type {1} however type {2} was registered.";
    t[3] = "CallableStatement функция бе обработена и изходния параметър {0} бе от тип {1}, обаче тип {2} бе използван.";
    t[6] = "Too many update results were returned.";
    t[7] = "Твърде много резултати бяха получени при актуализацията.";
    t[10] = "There are no rows in this ResultSet.";
    t[11] = "В този ResultSet няма редове.";
    t[14] = "Detail: {0}";
    t[15] = "Подробност: {0}";
    t[20] = "Invalid fetch direction constant: {0}.";
    t[21] = "Невалидна константа за fetch посоката: {0}.";
    t[22] = "No function outputs were registered.";
    t[23] = "Резултати от функцията не бяха регистрирани.";
    t[24] = "The array index is out of range: {0}";
    t[25] = "Индексът на масив е извън обхвата: {0}";
    t[26] = "The authentication type {0} is not supported. Check that you have configured the pg_hba.conf file to include the client''s IP address or subnet, and that it is using an authentication scheme supported by the driver.";
    t[27] = "Тип на удостоверяване {0} не се поддържа. Проверете дали сте конфигурирали pg_hba.conf файла, да включва IP адреса на клиента или подмрежата, и че се използва схема за удостоверяване, поддържана от драйвъра.";
    t[28] = "The server requested password-based authentication, but no password was provided.";
    t[29] = "Сървърът изисква идентифициране с парола, но парола не бе въведена.";
    t[40] = "Large Objects may not be used in auto-commit mode.";
    t[41] = "Големи обекти LOB не могат да се използват в auto-commit модус.";
    t[46] = "Operation requires a scrollable ResultSet, but this ResultSet is FORWARD_ONLY.";
    t[47] = "Операцията изисква резултатите да са scrollable, но този ResultSet е FORWARD_ONLY.";
    t[48] = "Zero bytes may not occur in string parameters.";
    t[49] = "Не може да има нула байта в низ параметрите.";
    t[50] = "The JVM claims not to support the encoding: {0}";
    t[51] = "JVM не поддържа тази кодова таблица за момента: {0}";
    t[54] = "Your security policy has prevented the connection from being attempted.  You probably need to grant the connect java.net.SocketPermission to the database server host and port that you wish to connect to.";
    t[55] = "Връзката не бе осъществена, поради вашите настройки за сигурност. Може би трябва да предоставите java.net.SocketPermission права на сървъра и порта с базата данни, към който искате да се свържете.";
    t[62] = "Database connection failed when canceling copy operation";
    t[63] = "Неосъществена връзка към базата данни при прекъсване на копирането";
    t[78] = "Error loading default settings from driverconfig.properties";
    t[79] = "Грешка при зареждане на настройките по подразбиране от файла driverconfig.properties";
    t[82] = "Returning autogenerated keys is not supported.";
    t[83] = "Автоматично генерирани ключове не се поддържат.";
    t[92] = "Unable to find name datatype in the system catalogs.";
    t[93] = "Не може да се намери името на типа данни в системните каталози.";
    t[94] = "Tried to read from inactive copy";
    t[95] = "Опит за четене при неактивно копиране";
    t[96] = "ResultSet is not updateable.  The query that generated this result set must select only one table, and must select all primary keys from that table. See the JDBC 2.1 API Specification, section 5.6 for more details.";
    t[97] = "ResultSet не може да се обновява. Заявката генерираща този резултат трябва да селектира само една таблица, както и всички първични ключове в нея. За повече информация, вижте раздел 5.6 на JDBC 2.1 API Specification.";
    t[98] = "Cannot cast an instance of {0} to type {1}";
    t[99] = "Не може да преобразува инстанция на {0} към тип {1}";
    t[102] = "Requested CopyOut but got {0}";
    t[103] = "Зададено CopyOut но получено {0}";
    t[106] = "Not implemented: Prepare must be issued using the same connection that started the transaction. currentXid={0}, prepare xid={1}";
    t[107] = "Невъзможна комбинация: Prepare трябва да бъде издадено чрез използване на същата връзка, при която е започната транзакцията. currentXid={0}, prepare xid={1}";
    t[108] = "Can''t use query methods that take a query string on a PreparedStatement.";
    t[109] = "Не може да се употребяват методи за заявка, които ползват низове на PreparedStatement.";
    t[114] = "Conversion of money failed.";
    t[115] = "Неуспешно валутно преобразуване.";
    t[118] = "Tried to obtain lock while already holding it";
    t[119] = "Опит за получаване на заключване/резервация докато вече е получено";
    t[120] = "This SQLXML object has not been initialized, so you cannot retrieve data from it.";
    t[121] = "Този SQLXML обект не е инициализиран, така че не могат да се извличат данни от него.";
    t[122] = "This SQLXML object has already been freed.";
    t[123] = "Този SQLXML обект вече е освободен.";
    t[124] = "Invalid stream length {0}.";
    t[125] = "Невалидна дължина {0} на потока данни.";
    t[130] = "Position: {0}";
    t[131] = "Позиция: {0}";
    t[132] = "The server does not support SSL.";
    t[133] = "Сървърът не поддържа SSL.";
    t[134] = "Got {0} error responses to single copy cancel request";
    t[135] = "Получени {0} отговори за грешка при единствено искане да се прекъсне копирането";
    t[136] = "DataSource has been closed.";
    t[137] = "Източникът на данни е прекъснат.";
    t[138] = "Unable to convert DOMResult SQLXML data to a string.";
    t[139] = "Не може да преобразува DOMResult SQLXML данни в низ.";
    t[144] = "Invalid UUID data.";
    t[145] = "Невалидни UUID данни.";
    t[148] = "The fastpath function {0} is unknown.";
    t[149] = "Функцията {0} е неизвестна.";
    t[154] = "Connection has been closed.";
    t[155] = "Връзката бе прекъсната.";
    t[156] = "This statement does not declare an OUT parameter.  Use '{' ?= call ... '}' to declare one.";
    t[157] = "Тази заявка не декларира изходен параметър. Ползвайте '{' ?= call ... '}' за да декларирате такъв.";
    t[158] = "A connection could not be made using the requested protocol {0}.";
    t[159] = "Не може да осъществи връзка, ползвайки искания протокол {0}.";
    t[162] = "The maximum field size must be a value greater than or equal to 0.";
    t[163] = "Максималният размер на полето трябва да бъде стойност по-голяма или равна на 0.";
    t[166] = "GSS Authentication failed";
    t[167] = "GSS удостоверяването бе неуспешно";
    t[176] = "Unknown XML Result class: {0}";
    t[177] = "Неизвестен XML изходящ клас: {0}";
    t[180] = "Server SQLState: {0}";
    t[181] = "SQL статус на сървъра: {0}";
    t[182] = "Unknown Response Type {0}.";
    t[183] = "Неизвестен тип на отговор {0}.";
    t[186] = "Tried to cancel an inactive copy operation";
    t[187] = "Опит за прекъсване на неактивно копиране";
    t[190] = "This PooledConnection has already been closed.";
    t[191] = "Тази PooledConnection връзка бе вече прекъсната.";
    t[200] = "Multiple ResultSets were returned by the query.";
    t[201] = "Заявката върна няколко ResultSets.";
    t[202] = "Finalizing a Connection that was never closed:";
    t[203] = "Приключване на връзка, която не бе прекъсната:";
    t[204] = "Unsupported Types value: {0}";
    t[205] = "Неподдържана стойност за тип: {0}";
    t[206] = "A CallableStatement was declared, but no call to registerOutParameter(1, <some type>) was made.";
    t[207] = "CallableStatement функция бе декларирана, но обработена като registerOutParameter(1, <some type>) ";
    t[208] = "Cannot retrieve the name of an unnamed savepoint.";
    t[209] = "Не може да определи името на неупомената savepoint.";
    t[220] = "Cannot change transaction read-only property in the middle of a transaction.";
    t[221] = "Не може да променяте правата на транзакцията по време на нейното извършване.";
    t[222] = "Bind message length {0} too long.  This can be caused by very large or incorrect length specifications on InputStream parameters.";
    t[223] = "Прекалено голяма дължина {0} на съобщението. Това може да е причинено от прекалено голяма или неправилно зададена дължина на InputStream параметри.";
    t[224] = "The parameter index is out of range: {0}, number of parameters: {1}.";
    t[225] = "Параметърният индекс е извън обхват: {0}, брой параметри: {1}.";
    t[226] = "Transaction isolation level {0} not supported.";
    t[227] = "Изолационно ниво на транзакциите {0} не се поддържа.";
    t[234] = "Cannot update the ResultSet because it is either before the start or after the end of the results.";
    t[235] = "Не може да се обнови ResultSet, когато се намираме преди началото или след края на резултатите.";
    t[238] = "tried to call end without corresponding start call. state={0}, start xid={1}, currentXid={2}, preparedXid={3}";
    t[239] = "опита да извика end без съответстващо извикване на start. state={0}, start xid={1}, currentXid={2}, preparedXid={3}";
    t[242] = "This SQLXML object has already been initialized, so you cannot manipulate it further.";
    t[243] = "Този SQLXML обект вече е инициализиран и не може да бъде променен.";
    t[250] = "Conversion to type {0} failed: {1}.";
    t[251] = "Неуспешно преобразуване към тип {0}: {1}.";
    t[252] = "The SSLSocketFactory class provided {0} could not be instantiated.";
    t[253] = "Класът SSLSocketFactory връща {0} и не може да бъде инстанцииран.";
    t[254] = "Unable to create SAXResult for SQLXML.";
    t[255] = "Не може да се създаде SAXResult за SQLXML.";
    t[256] = "Interrupted while attempting to connect.";
    t[257] = "Опита за осъществяване на връзка бе своевременно прекъснат. ";
    t[260] = "Protocol error.  Session setup failed.";
    t[261] = "Грешка в протокола. Неуспешна настройка на сесията.";
    t[264] = "Database connection failed when starting copy";
    t[265] = "Неосъществена връзка към базата данни при започване на копирането";
    t[272] = "Cannot call cancelRowUpdates() when on the insert row.";
    t[273] = "Не може да се изпълни cancelRowUpdates() метода, когато се намираме при редицата на въвеждане.";
    t[274] = "Unable to bind parameter values for statement.";
    t[275] = "Не може да подготви параметрите на командата.";
    t[280] = "A result was returned when none was expected.";
    t[281] = "Бе получен резултат, когато такъв не бе очакван.";
    t[282] = "The server''s standard_conforming_strings parameter was reported as {0}. The JDBC driver expected on or off.";
    t[283] = "Параметърът standard_conforming_strings при сървъра бе докладван като {0}. JDBC драйвъра очаква този параметър да бъде on или off.";
    t[284] = "Unable to translate data into the desired encoding.";
    t[285] = "Невъзможно преобразуване на данни в желаното кодиране.";
    t[292] = "PostgreSQL LOBs can only index to: {0}";
    t[293] = "PostgreSQL индексира големи обекти LOB само до: {0}";
    t[294] = "Provided InputStream failed.";
    t[295] = "Зададения InputStream поток е неуспешен.";
    t[296] = "Invalid protocol state requested. Attempted transaction interleaving is not supported. xid={0}, currentXid={1}, state={2}, flags={3}";
    t[297] = "Транзакция в транзакция не се поддържа за момента. xid={0}, currentXid={1}, state={2}, flags={3}";
    t[304] = "{0} function takes four and only four argument.";
    t[305] = "Функцията {0} може да приеме четири и само четири аргумента.";
    t[306] = "{0} function doesn''t take any argument.";
    t[307] = "Функцията {0} не може да приема аргументи.";
    t[310] = "Got CopyOutResponse from server during an active {0}";
    t[311] = "Получен CopyOutResponse отговор от сървъра при активно {0}";
    t[322] = "No value specified for parameter {0}.";
    t[323] = "Няма стойност, определена за параметър {0}.";
    t[324] = "Illegal UTF-8 sequence: initial byte is {0}: {1}";
    t[325] = "Невалидна UTF-8 последователност: първоначален байт е {0}: {1}";
    t[326] = "Error disabling autocommit";
    t[327] = "Грешка при изключване на autocommit";
    t[328] = "Illegal UTF-8 sequence: byte {0} of {1} byte sequence is not 10xxxxxx: {2}";
    t[329] = "Невалидна UTF-8 последователност: байта {0} от байтова последователност {1} не е 10xxxxxx: {2}";
    t[330] = "Received CommandComplete ''{0}'' without an active copy operation";
    t[331] = "Получено командно допълнение ''{0}'' без активна команда за копиране";
    t[332] = "Illegal UTF-8 sequence: final value is out of range: {0}";
    t[333] = "Невалидна UTF-8 последователност: крайната стойност е извън стойностните граници: {0}";
    t[336] = "Cannot change transaction isolation level in the middle of a transaction.";
    t[337] = "Не може да променяте изолационното ниво на транзакцията по време на нейното извършване.";
    t[340] = "An unexpected result was returned by a query.";
    t[341] = "Заявката върна неочакван резултат.";
    t[346] = "Conversion of interval failed";
    t[347] = "Неуспешно преобразуване на интервал";
    t[350] = "This ResultSet is closed.";
    t[351] = "Операциите по този ResultSet са били прекратени.";
    t[352] = "Read from copy failed.";
    t[353] = "Четене от копието неуспешно.";
    t[354] = "Unable to load the class {0} responsible for the datatype {1}";
    t[355] = "Невъзможно е зареждането на клас {0}, отговарящ за типа данни {1}";
    t[356] = "Failed to convert binary xml data to encoding: {0}.";
    t[357] = "Неуспешно преобразуване на двоични XML данни за кодиране съгласно: {0}.";
    t[362] = "Connection attempt timed out.";
    t[363] = "Времето за осъществяване на връзката изтече (таймаут).";
    t[364] = "Expected command status BEGIN, got {0}.";
    t[365] = "Очаквана команда BEGIN, получена {0}.";
    t[372] = "This copy stream is closed.";
    t[373] = "Потока за копиране на данните е затворен.";
    t[376] = "Can''t infer the SQL type to use for an instance of {0}. Use setObject() with an explicit Types value to specify the type to use.";
    t[377] = "Не може да се определи SQL тип, който да се използва за инстанцията на {0}. Ползвайте метода setObject() с точни стойности, за да определите типа.";
    t[378] = "Can''t refresh the insert row.";
    t[379] = "Не може да обнови въведения ред.";
    t[382] = "You must specify at least one column value to insert a row.";
    t[383] = "Трябва да посочите поне една стойност за колона, за да вмъкнете ред.";
    t[388] = "Connection is busy with another transaction";
    t[389] = "Връзката е заета с друга транзакция";
    t[392] = "Bad value for type {0} : {1}";
    t[393] = "Невалидна стойност за тип {0} : {1}";
    t[396] = "This statement has been closed.";
    t[397] = "Командата е извършена.";
    t[404] = "No primary key found for table {0}.";
    t[405] = "Няма първичен ключ за таблица {0}.";
    t[406] = "Currently positioned after the end of the ResultSet.  You cannot call deleteRow() here.";
    t[407] = "В момента се намираме преди края на ResultSet. Тук не може да се изпълни deleteRow() метода.";
    t[414] = "{0} function takes two or three arguments.";
    t[415] = "Функцията {0} може да приеме два или три аргумента.";
    t[416] = "{0} function takes three and only three arguments.";
    t[417] = "Функцията {0} може да приеме три и само три аргумента.";
    t[418] = "Unable to find server array type for provided name {0}.";
    t[419] = "Не може да се намери типа на сървърен масив за зададеното име {0}.";
    t[420] = "Fastpath call {0} - No result was returned and we expected an integer.";
    t[421] = "Извикване на {0} - няма резултати и а бе очаквано цяло число.";
    t[426] = "Database connection failed when ending copy";
    t[427] = "Неосъществена връзка към базата данни при завършване на копирането";
    t[428] = "Cannot write to copy a byte of value {0}";
    t[429] = "Няма пишещи права, за да копира байтова стойност {0}";
    t[430] = "Results cannot be retrieved from a CallableStatement before it is executed.";
    t[431] = "Резултати от CallableStatement функция не могат да бъдат получени, преди тя да бъде обработена.";
    t[432] = "Cannot reference a savepoint after it has been released.";
    t[433] = "Не може да референцира savepoint, след като е била освободена.";
    t[434] = "Failed to create object for: {0}.";
    t[435] = "Неуспешно създаване на обект за: {0}.";
    t[438] = "Unexpected packet type during copy: {0}";
    t[439] = "Неочакван тип пакет при копиране: {0}";
    t[442] = "Unable to determine a value for MaxIndexKeys due to missing system catalog data.";
    t[443] = "Невъзможно е да се определи стойността за MaxIndexKeys поради липса на системния каталог с данни.";
    t[444] = "Tried to end inactive copy";
    t[445] = "Опит за прекъсване на неактивно копиране";
    t[450] = "Unexpected copydata from server for {0}";
    t[451] = "Неочаквано CopyData от сървъра за {0}";
    t[460] = "Zero bytes may not occur in identifiers.";
    t[461] = "Не може да има нула байта в идентификаторите.";
    t[462] = "Error during one-phase commit. commit xid={0}";
    t[463] = "Грешка при едно-фазов commit. commit xid={0}";
    t[464] = "Ran out of memory retrieving query results.";
    t[465] = "Недостатъчна памет при представяна на резултатите от заявката.";
    t[468] = "Unable to create StAXResult for SQLXML";
    t[469] = "Не може да се създаде StAXResult за SQLXML.";
    t[470] = "Location: File: {0}, Routine: {1}, Line: {2}";
    t[471] = "Местоположение: Файл: {0}, Функция: {1}, Ред: {2}";
    t[482] = "A CallableStatement was executed with an invalid number of parameters";
    t[483] = "CallableStatement функция бе обработена, но с непозволен брой параметри.";
    t[486] = "Illegal UTF-8 sequence: {0} bytes used to encode a {1} byte value: {2}";
    t[487] = "Невалидна UTF-8 последователност: {0} байта използвани за кодирането на {1} байтова стойност: {2}";
    t[496] = "Interrupted while waiting to obtain lock on database connection";
    t[497] = "Прекъсване при чакане да получи заключване/резервация при връзка към базата данни";
    t[502] = "LOB positioning offsets start at 1.";
    t[503] = "Позиционалният офсет при големи обекти LOB започва от 1.";
    t[506] = "Returning autogenerated keys by column index is not supported.";
    t[507] = "Автоматично генерирани ключове спрямо индекс на колона не се поддържат.";
    t[510] = "Currently positioned before the start of the ResultSet.  You cannot call deleteRow() here.";
    t[511] = "В момента се намираме в началото на ResultSet. Тук не може да се изпълни deleteRow() метода.";
    t[524] = "Truncation of large objects is only implemented in 8.3 and later servers.";
    t[525] = "Скъсяване на големи обекти LOB е осъществено само във версии след 8.3.";
    t[526] = "Statement has been closed.";
    t[527] = "Командата е завършена.";
    t[540] = "Database connection failed when writing to copy";
    t[541] = "Неосъществена връзка към базата данни при опит за копиране";
    t[544] = "The server''s DateStyle parameter was changed to {0}. The JDBC driver requires DateStyle to begin with ISO for correct operation.";
    t[545] = "Параметърът DateStyle при сървъра бе променен на {0}. JDBC драйвъра изисква DateStyle започва с ISO за да функционира правилно.";
    t[546] = "Provided Reader failed.";
    t[547] = "Грешка с ползвания четец.";
    t[550] = "Not on the insert row.";
    t[551] = "Не сме в редицата на въвеждане.";
    t[566] = "Unable to decode xml data.";
    t[567] = "Не може да декодира XML данните.";
    t[570] = "Not implemented: 2nd phase commit must be issued using an idle connection. commit xid={0}, currentXid={1}, state={2}, transactionState={3}";
    t[571] = "Невъзможна комбинация: втората фаза на commit задължително трябва да бъде издадена при свободна връзка. commit xid={0}, currentXid={1}, state={2}, transactionState={3}";
    t[596] = "Tried to write to an inactive copy operation";
    t[597] = "Опит за писане при неактивна операция за копиране";
    t[606] = "An error occurred while setting up the SSL connection.";
    t[607] = "Възникна грешка при осъществяване на SSL връзката.";
    t[614] = "Something unusual has occurred to cause the driver to fail. Please report this exception.";
    t[615] = "Възникна неочаквана грешка с драйвъра. Моля докадвайте това изключение. ";
    t[618] = "No results were returned by the query.";
    t[619] = "Няма намерени резултати за заявката.";
    t[620] = "ClientInfo property not supported.";
    t[621] = "Информацията за ClientInfo не се поддържа.";
    t[622] = "Unexpected error writing large object to database.";
    t[623] = "Неочаквана грешка при записване на голям обект LOB в базата данни.";
    t[628] = "The JVM claims not to support the {0} encoding.";
    t[629] = "JVM не поддържа за момента {0} кодовата таблица.";
    t[630] = "Unknown XML Source class: {0}";
    t[631] = "Неизвестен XML входящ клас: {0}";
    t[632] = "Interval {0} not yet implemented";
    t[633] = "Интервалът {0} не е валиден все още.";
    t[636] = "commit called before end. commit xid={0}, state={1}";
    t[637] = "commit извикан преди end. commit xid={0}, state={1}";
    t[638] = "Tried to break lock on database connection";
    t[639] = "Опит за премахване на заключването/резервацията при връзка към базата данни";
    t[642] = "Missing expected error response to copy cancel request";
    t[643] = "Липсва очакван отговор при грешка да прекъсне копирането";
    t[644] = "Maximum number of rows must be a value grater than or equal to 0.";
    t[645] = "Максималният брой редове трябва да бъде стойност по-голяма или равна на 0.";
    t[652] = "Requested CopyIn but got {0}";
    t[653] = "Зададено CopyIn но получено {0}";
    t[656] = "Parameter of type {0} was registered, but call to get{1} (sqltype={2}) was made.";
    t[657] = "Отчетен параметър от тип {0}, но обработено като get{1} (sqltype={2}). ";
    t[662] = "Unsupported value for stringtype parameter: {0}";
    t[663] = "Непозволена стойност за StringType параметър: {0}";
    t[664] = "Fetch size must be a value greater than or equal to 0.";
    t[665] = "Размера за fetch size трябва да бъде по-голям или равен на 0.";
    t[670] = "Cannot tell if path is open or closed: {0}.";
    t[671] = "Не може да определи дали адреса е отворен или затворен: {0}.";
    t[672] = "Expected an EOF from server, got: {0}";
    t[673] = "Очакван край на файла от сървъра, но получено: {0}";
    t[680] = "Copying from database failed: {0}";
    t[681] = "Копирането от базата данни бе неуспешно: {0}";
    t[682] = "Connection has been closed automatically because a new connection was opened for the same PooledConnection or the PooledConnection has been closed.";
    t[683] = "Връзката бе автоматично прекъсната, защото нова връзка за същата беше осъществена или PooledConnection връзката е вече прекъсната.";
    t[698] = "Custom type maps are not supported.";
    t[699] = "Специфични типови съответствия не се поддържат.";
    t[700] = "xid must not be null";
    t[701] = "xid не може да бъде null";
    t[706] = "Internal Position: {0}";
    t[707] = "Вътрешна позиция: {0}";
    t[708] = "Error during recover";
    t[709] = "Грешка при възстановяване";
    t[712] = "Method {0} is not yet implemented.";
    t[713] = "Методът {0} все още не е функционален.";
    t[714] = "Unexpected command status: {0}.";
    t[715] = "Неочакван статус на команда: {0}.";
    t[718] = "The column index is out of range: {0}, number of columns: {1}.";
    t[719] = "Индексът на колоната е извън стойностен обхват: {0}, брой колони: {1}.";
    t[730] = "Unknown ResultSet holdability setting: {0}.";
    t[731] = "Неизвестна ResultSet holdability настройка: {0}.";
    t[734] = "Cannot call deleteRow() when on the insert row.";
    t[735] = "Не може да се изпълни deleteRow() метода, когато се намираме при редицата на въвеждане.";
    t[740] = "ResultSet not positioned properly, perhaps you need to call next.";
    t[741] = "ResultSet не е референциран правилно. Вероятно трябва да придвижите курсора посредством next.";
    t[742] = "wasNull cannot be call before fetching a result.";
    t[743] = "wasNull не може да бьде изпълнен, преди наличието на резултата.";
    t[746] = "{0} function takes two and only two arguments.";
    t[747] = "Функцията {0} може да приеме два и само два аргумента.";
    t[750] = "Malformed function or procedure escape syntax at offset {0}.";
    t[751] = "Непозволен синтаксис на функция или процедура при офсет {0}.";
    t[752] = "Premature end of input stream, expected {0} bytes, but only read {1}.";
    t[753] = "Преждевременен край на входящ поток на данни, очаквани {0} байта, но прочетени само {1}.";
    t[756] = "Got CopyData without an active copy operation";
    t[757] = "Получено CopyData без наличие на активна операция за копиране";
    t[758] = "Cannot retrieve the id of a named savepoint.";
    t[759] = "Не може да определи ID на спомената savepoint.";
    t[770] = "Where: {0}";
    t[771] = "Където: {0}";
    t[778] = "Got CopyInResponse from server during an active {0}";
    t[779] = "Получен CopyInResponse отговор от сървъра при активно {0}";
    t[780] = "Cannot convert an instance of {0} to type {1}";
    t[781] = "Не може да преобразува инстанцията на {0} във вида {1}";
    t[784] = "Not implemented: one-phase commit must be issued using the same connection that was used to start it";
    t[785] = "Невъзможна комбинация: едно-фазов commit трябва да бъде издаден чрез използване на същата връзка, при която е започнал";
    t[790] = "Invalid flags {0}";
    t[791] = "Невалидни флагове {0}";
    t[798] = "Query timeout must be a value greater than or equals to 0.";
    t[799] = "Времето за изпълнение на заявката трябва да бъде стойност по-голяма или равна на 0.";
    t[802] = "Hint: {0}";
    t[803] = "Забележка: {0}";
    t[810] = "The array index is out of range: {0}, number of elements: {1}.";
    t[811] = "Индексът на масив е извън обхвата: {0}, брой елементи: {1}.";
    t[812] = "Internal Query: {0}";
    t[813] = "Вътрешна заявка: {0}";
    t[816] = "CommandComplete expected COPY but got: ";
    t[817] = "Очаквано командно допълнение COPY но получено: ";
    t[824] = "Illegal UTF-8 sequence: final value is a surrogate value: {0}";
    t[825] = "Невалидна UTF-8 последователност: крайната стойност е заместителна стойност: {0}";
    t[826] = "Unknown type {0}.";
    t[827] = "Неизвестен тип {0}.";
    t[828] = "ResultSets with concurrency CONCUR_READ_ONLY cannot be updated.";
    t[829] = "ResultSets с concurrency CONCUR_READ_ONLY не могат да бъдат актуализирани.";
    t[830] = "The connection attempt failed.";
    t[831] = "Опита за връзка бе неуспешен.";
    t[834] = "{0} function takes one and only one argument.";
    t[835] = "Функцията {0} може да приеме само един единствен аргумент.";
    t[838] = "suspend/resume not implemented";
    t[839] = "спиране / започване не се поддържа за момента";
    t[840] = "Error preparing transaction. prepare xid={0}";
    t[841] = "Грешка при подготвяне на транзакция. prepare xid={0}";
    t[842] = "The driver currently does not support COPY operations.";
    t[843] = "За момента драйвъра не поддържа COPY команди.";
    t[852] = "Heuristic commit/rollback not supported. forget xid={0}";
    t[853] = "Евристичен commit или rollback не се поддържа. forget xid={0}";
    t[856] = "Invalid character data was found.  This is most likely caused by stored data containing characters that are invalid for the character set the database was created in.  The most common example of this is storing 8bit data in a SQL_ASCII database.";
    t[857] = "Бяха намерени невалидни данни. Това най-вероятно се дължи на съхранявани данни, съдържащи символи, които са невалидни за набора от знаци при създаване на базата данни. Чест пример за това е съхраняване на 8bit данни в SQL_ASCII бази данни.";
    t[858] = "Cannot establish a savepoint in auto-commit mode.";
    t[859] = "Не може да се установи savepoint в auto-commit модус.";
    t[862] = "The column name {0} was not found in this ResultSet.";
    t[863] = "Името на колоната {0} не бе намерено в този ResultSet.";
    t[864] = "Prepare called before end. prepare xid={0}, state={1}";
    t[865] = "Prepare извикано преди края. prepare xid={0}, state={1}";
    t[866] = "Unknown Types value.";
    t[867] = "Стойност от неизвестен тип.";
    t[870] = "Cannot call updateRow() when on the insert row.";
    t[871] = "Не може да се изпълни updateRow() метода, когато се намираме при редицата на въвеждане.";
    t[876] = "Database connection failed when reading from copy";
    t[877] = "Неосъществена връзка към базата данни при четене от копие";
    t[880] = "Error rolling back prepared transaction. rollback xid={0}, preparedXid={1}, currentXid={2}";
    t[881] = "Грешка при възстановяване на състоянието преди подготвена транзакция. rollback xid={0}, preparedXid={1}, currentXid={2}";
    t[882] = "Can''t use relative move methods while on the insert row.";
    t[883] = "Не може да се използват относителни методи за движение, когато се намираме при редицата на въвеждане.";
    t[884] = "free() was called on this LOB previously";
    t[885] = "Функцията free() бе вече извикана за този голям обект LOB";
    t[888] = "A CallableStatement was executed with nothing returned.";
    t[889] = "CallableStatement функция бе обработена, но няма резултати.";
    table = t;
  }
  public java.lang.Object handleGetObject (java.lang.String msgid) throws java.util.MissingResourceException {
    int hash_val = msgid.hashCode() & 0x7fffffff;
    int idx = (hash_val % 445) << 1;
    {
      java.lang.Object found = table[idx];
      if (found == null)
        return null;
      if (msgid.equals(found))
        return table[idx + 1];
    }
    int incr = ((hash_val % 443) + 1) << 1;
    for (;;) {
      idx += incr;
      if (idx >= 890)
        idx -= 890;
      java.lang.Object found = table[idx];
      if (found == null)
        return null;
      if (msgid.equals(found))
        return table[idx + 1];
    }
  }
  public java.util.Enumeration getKeys () {
    return
      new java.util.Enumeration() {
        private int idx = 0;
        { while (idx < 890 && table[idx] == null) idx += 2; }
        public boolean hasMoreElements () {
          return (idx < 890);
        }
        public java.lang.Object nextElement () {
          java.lang.Object key = table[idx];
          do idx += 2; while (idx < 890 && table[idx] == null);
          return key;
        }
      };
  }
  public java.util.ResourceBundle getParent () {
    return parent;
  }
}
