/* Automatically generated by GNU msgfmt.  Do not modify!  */
package org.postgresql.translation;
public class messages_cs extends java.util.ResourceBundle {
  private static final java.lang.String[] table;
  static {
    java.lang.String[] t = new java.lang.String[314];
    t[0] = "";
    t[1] = "Project-Id-Version: PostgreSQL JDBC Driver 8.0\nReport-Msgid-Bugs-To: \nPO-Revision-Date: 2005-08-21 20:00+0200\nLast-Translator: Petr Dittrich <bodyn@medoro.org>\nLanguage-Team: \nLanguage: \nMIME-Version: 1.0\nContent-Type: text/plain; charset=UTF-8\nContent-Transfer-Encoding: 8bit\n";
    t[2] = "A connection could not be made using the requested protocol {0}.";
    t[3] = "Spojení nelze vytvořit s použitím žádaného protokolu {0}.";
    t[4] = "Malformed function or procedure escape syntax at offset {0}.";
    t[5] = "Poškozená funkce nebo opuštění procedury na pozici {0}.";
    t[8] = "Cannot cast an instance of {0} to type {1}";
    t[9] = "Nemohu přetypovat instanci {0} na typ {1}";
    t[12] = "ResultSet is not updateable.  The query that generated this result set must select only one table, and must select all primary keys from that table. See the JDBC 2.1 API Specification, section 5.6 for more details.";
    t[13] = "ResultSet není aktualizavatelný. Dotaz musí vybírat pouze z jedné tabulky a musí obsahovat všechny primární klíče tabulky. Koukni do JDBC 2.1 API Specifikace, sekce 5.6 pro více podrobností.";
    t[14] = "The JVM claims not to support the {0} encoding.";
    t[15] = "JVM tvrdí, že nepodporuje kodování {0}.";
    t[16] = "An I/O error occurred while sending to the backend.";
    t[17] = "Vystupně/výstupní chyba při odesílání k backend.";
    t[18] = "Statement has been closed.";
    t[19] = "Statement byl uzavřen.";
    t[20] = "Unknown Types value.";
    t[21] = "Neznámá hodnota typu.";
    t[22] = "ResultSets with concurrency CONCUR_READ_ONLY cannot be updated.";
    t[23] = "ResultSets se souběžností CONCUR_READ_ONLY nemůže být aktualizováno";
    t[26] = "You must specify at least one column value to insert a row.";
    t[27] = "Musíte vyplnit alespoň jeden sloupec pro vložení řádku.";
    t[32] = "No primary key found for table {0}.";
    t[33] = "Nenalezen primární klíč pro tabulku {0}.";
    t[34] = "Cannot establish a savepoint in auto-commit mode.";
    t[35] = "Nemohu vytvořit savepoint v auto-commit modu.";
    t[38] = "Can''t use relative move methods while on the insert row.";
    t[39] = "Nemůžete používat relativní přesuny při vkládání řádku.";
    t[44] = "The column name {0} was not found in this ResultSet.";
    t[45] = "Sloupec pojmenovaný {0} nebyl nalezen v ResultSet.";
    t[46] = "This statement has been closed.";
    t[47] = "Příkaz byl uzavřen.";
    t[48] = "The SSLSocketFactory class provided {0} could not be instantiated.";
    t[49] = "Třída SSLSocketFactory poskytla {0} což nemůže být instancionizováno.";
    t[50] = "Multiple ResultSets were returned by the query.";
    t[51] = "Vícenásobný ResultSet byl vrácen dotazem.";
    t[52] = "DataSource has been closed.";
    t[53] = "DataSource byl uzavřen.";
    t[56] = "Error loading default settings from driverconfig.properties";
    t[57] = "Chyba načítání standardního nastavení z driverconfig.properties";
    t[62] = "Bad value for type {0} : {1}";
    t[63] = "Špatná hodnota pro typ {0} : {1}";
    t[66] = "Method {0} is not yet implemented.";
    t[67] = "Metoda {0} není implementována.";
    t[68] = "The array index is out of range: {0}";
    t[69] = "Index pole mimo rozsah: {0}";
    t[70] = "Unexpected command status: {0}.";
    t[71] = "Neočekávaný stav příkazu: {0}.";
    t[74] = "Expected command status BEGIN, got {0}.";
    t[75] = "Očekáván příkaz BEGIN, obdržen {0}.";
    t[76] = "Cannot retrieve the id of a named savepoint.";
    t[77] = "Nemohu získat id nepojmenovaného savepointu.";
    t[78] = "Unexpected error writing large object to database.";
    t[79] = "Neočekávaná chyba při zapisování velkého objektu do databáze.";
    t[84] = "Not on the insert row.";
    t[85] = "Ne na vkládaném řádku.";
    t[86] = "Returning autogenerated keys is not supported.";
    t[87] = "Vrácení automaticky generovaných klíčů není podporováno.";
    t[88] = "The server requested password-based authentication, but no password was provided.";
    t[89] = "Server vyžaduje ověření heslem, ale žádné nebylo posláno.";
    t[98] = "Unable to load the class {0} responsible for the datatype {1}";
    t[99] = "Nemohu načíst třídu {0} odpovědnou za typ {1}";
    t[100] = "Invalid fetch direction constant: {0}.";
    t[101] = "Špatný směr čtení: {0}.";
    t[102] = "Conversion of money failed.";
    t[103] = "Převod peněz selhal.";
    t[104] = "Connection has been closed.";
    t[105] = "Spojeni bylo uzavřeno.";
    t[106] = "Cannot retrieve the name of an unnamed savepoint.";
    t[107] = "Nemohu získat název nepojmenovaného savepointu.";
    t[108] = "Large Objects may not be used in auto-commit mode.";
    t[109] = "Velké objecky nemohou být použity v auto-commit modu.";
    t[110] = "This ResultSet is closed.";
    t[111] = "Tento ResultSet je uzavřený.";
    t[116] = "Something unusual has occurred to cause the driver to fail. Please report this exception.";
    t[117] = "Něco neobvyklého přinutilo ovladač selhat. Prosím nahlaste tuto vyjímku.";
    t[118] = "The server does not support SSL.";
    t[119] = "Server nepodporuje SSL.";
    t[120] = "Invalid stream length {0}.";
    t[121] = "Vadná délka proudu {0}.";
    t[126] = "The maximum field size must be a value greater than or equal to 0.";
    t[127] = "Maximální velikost pole musí být nezáporné číslo.";
    t[130] = "Cannot call updateRow() when on the insert row.";
    t[131] = "Nemohu volat updateRow() na vlkádaném řádku.";
    t[132] = "A CallableStatement was executed with nothing returned.";
    t[133] = "CallableStatement byl spuštěn, leč nic nebylo vráceno.";
    t[134] = "Provided Reader failed.";
    t[135] = "Selhal poskytnutý Reader.";
    t[146] = "Cannot call deleteRow() when on the insert row.";
    t[147] = "Nemůžete volat deleteRow() při vkládání řádku.";
    t[156] = "Where: {0}";
    t[157] = "Kde: {0}";
    t[158] = "An unexpected result was returned by a query.";
    t[159] = "Obdržen neočekávaný výsledek dotazu.";
    t[160] = "The connection attempt failed.";
    t[161] = "Pokus o připojení selhal.";
    t[162] = "Too many update results were returned.";
    t[163] = "Bylo vráceno příliš mnoho výsledků aktualizací.";
    t[164] = "Unknown type {0}.";
    t[165] = "Neznámý typ {0}.";
    t[166] = "{0} function takes two and only two arguments.";
    t[167] = "Funkce {0} bere právě dva argumenty.";
    t[168] = "{0} function doesn''t take any argument.";
    t[169] = "Funkce {0} nebere žádný argument.";
    t[172] = "Unable to find name datatype in the system catalogs.";
    t[173] = "Nemohu najít název typu v systémovém katalogu.";
    t[174] = "Protocol error.  Session setup failed.";
    t[175] = "Chyba protokolu. Nastavení relace selhalo.";
    t[176] = "{0} function takes one and only one argument.";
    t[177] = "Funkce {0} bere jeden argument.";
    t[186] = "The driver currently does not support COPY operations.";
    t[187] = "Ovladač nyní nepodporuje příkaz COPY.";
    t[190] = "Invalid character data was found.  This is most likely caused by stored data containing characters that are invalid for the character set the database was created in.  The most common example of this is storing 8bit data in a SQL_ASCII database.";
    t[191] = "Nalezena vada ve znakových datech. Toto může být způsobeno uloženými daty obsahujícími znaky, které jsou závadné pro znakovou sadu nastavenou při zakládání databáze. Nejznámejší příklad je ukládání 8bitových dat vSQL_ASCII databázi.";
    t[196] = "Fetch size must be a value greater than or equal to 0.";
    t[197] = "Nabraná velikost musí být nezáporná.";
    t[204] = "Unsupported Types value: {0}";
    t[205] = "Nepodporovaná hodnota typu: {0}";
    t[206] = "Can''t refresh the insert row.";
    t[207] = "Nemohu obnovit vkládaný řádek.";
    t[210] = "Maximum number of rows must be a value grater than or equal to 0.";
    t[211] = "Maximální počet řádek musí být nezáporné číslo.";
    t[216] = "No value specified for parameter {0}.";
    t[217] = "Nespecifikována hodnota parametru {0}.";
    t[218] = "The array index is out of range: {0}, number of elements: {1}.";
    t[219] = "Index pole mimo rozsah: {0}, počet prvků: {1}.";
    t[220] = "Provided InputStream failed.";
    t[221] = "Selhal poskytnutý InputStream.";
    t[228] = "Cannot reference a savepoint after it has been released.";
    t[229] = "Nemohu získat odkaz na savepoint, když byl uvolněn.";
    t[232] = "An error occurred while setting up the SSL connection.";
    t[233] = "Nastala chyba při nastavení SSL spojení.";
    t[246] = "Detail: {0}";
    t[247] = "Detail: {0}";
    t[248] = "This PooledConnection has already been closed.";
    t[249] = "Tento PooledConnection byl uzavřen.";
    t[250] = "A result was returned when none was expected.";
    t[251] = "Obdržen výsledek, ikdyž žádný nebyl očekáván.";
    t[254] = "The JVM claims not to support the encoding: {0}";
    t[255] = "JVM tvrdí, že nepodporuje kodování: {0}";
    t[256] = "The parameter index is out of range: {0}, number of parameters: {1}.";
    t[257] = "Index parametru mimo rozsah: {0}, počet parametrů {1}.";
    t[258] = "LOB positioning offsets start at 1.";
    t[259] = "Začátek pozicování LOB začína na 1.";
    t[260] = "{0} function takes two or three arguments.";
    t[261] = "Funkce {0} bere dva nebo tři argumenty.";
    t[262] = "Currently positioned after the end of the ResultSet.  You cannot call deleteRow() here.";
    t[263] = "Právě jste za pozicí konce ResultSetu. Zde nemůžete volat deleteRow().s";
    t[266] = "Server SQLState: {0}";
    t[267] = "Server SQLState: {0}";
    t[270] = "{0} function takes four and only four argument.";
    t[271] = "Funkce {0} bere přesně čtyři argumenty.";
    t[272] = "Failed to create object for: {0}.";
    t[273] = "Selhalo vytvoření objektu: {0}.";
    t[274] = "No results were returned by the query.";
    t[275] = "Neobdržen žádný výsledek dotazu.";
    t[276] = "Position: {0}";
    t[277] = "Pozice: {0}";
    t[278] = "The column index is out of range: {0}, number of columns: {1}.";
    t[279] = "Index sloupece je mimo rozsah: {0}, počet sloupců: {1}.";
    t[280] = "Unknown Response Type {0}.";
    t[281] = "Neznámý typ odpovědi {0}.";
    t[284] = "Hint: {0}";
    t[285] = "Rada: {0}";
    t[286] = "Location: File: {0}, Routine: {1}, Line: {2}";
    t[287] = "Poloha: Soubor: {0}, Rutina: {1}, Řádek: {2}";
    t[288] = "Query timeout must be a value greater than or equals to 0.";
    t[289] = "Časový limit dotazu musí být nezáporné číslo.";
    t[292] = "Unable to translate data into the desired encoding.";
    t[293] = "Nemohu přeložit data do požadovaného kódování.";
    t[296] = "Cannot call cancelRowUpdates() when on the insert row.";
    t[297] = "Nemůžete volat cancelRowUpdates() při vkládání řádku.";
    t[298] = "The authentication type {0} is not supported. Check that you have configured the pg_hba.conf file to include the client''s IP address or subnet, and that it is using an authentication scheme supported by the driver.";
    t[299] = "Ověření typu {0} není podporováno. Zkontrolujte zda konfigurační soubor pg_hba.conf obsahuje klientskou IP adresu či podsíť a zda je použité ověřenovací schéma podporováno ovladačem.";
    t[308] = "There are no rows in this ResultSet.";
    t[309] = "Žádný řádek v ResultSet.";
    table = t;
  }
  public java.lang.Object handleGetObject (java.lang.String msgid) throws java.util.MissingResourceException {
    int hash_val = msgid.hashCode() & 0x7fffffff;
    int idx = (hash_val % 157) << 1;
    {
      java.lang.Object found = table[idx];
      if (found == null)
        return null;
      if (msgid.equals(found))
        return table[idx + 1];
    }
    int incr = ((hash_val % 155) + 1) << 1;
    for (;;) {
      idx += incr;
      if (idx >= 314)
        idx -= 314;
      java.lang.Object found = table[idx];
      if (found == null)
        return null;
      if (msgid.equals(found))
        return table[idx + 1];
    }
  }
  public java.util.Enumeration getKeys () {
    return
      new java.util.Enumeration() {
        private int idx = 0;
        { while (idx < 314 && table[idx] == null) idx += 2; }
        public boolean hasMoreElements () {
          return (idx < 314);
        }
        public java.lang.Object nextElement () {
          java.lang.Object key = table[idx];
          do idx += 2; while (idx < 314 && table[idx] == null);
          return key;
        }
      };
  }
  public java.util.ResourceBundle getParent () {
    return parent;
  }
}
