/*
 * Copyright (c) 2020, PostgreSQL Global Development Group
 * See the LICENSE file in the project root for more information.
 */

package org.postgresql.test;

/**
 * Declares interface to specify slow tests.
 */
public interface SlowTests {
}
