/*
 * Copyright (c) 2004, PostgreSQL Global Development Group
 * See the LICENSE file in the project root for more information.
 */

package org.postgresql.jdbc2.optional;

import org.postgresql.ds.PGConnectionPoolDataSource;

/**
 * @deprecated Please use {@link PGConnectionPoolDataSource}
 */
@Deprecated
public class ConnectionPool extends PGConnectionPoolDataSource {
}
